@@ -14,36 +14,24 @@ try:
     import select
 except ImportError:
     select = None
+import re
 import socket
 import time
 
 import six
 from six import b
 
-from .subject_alt_name import get_subject_alt_name
+try:
+    from .subject_alt_name import get_subject_alt_name
+except ImportError:
+    get_subject_alt_name = None
 
-
-class _FailOnAccess(object):
-    def __init__(self, error):
-        self.error = error
-
-    def __getattribute__(self, name):
-        raise self.error
-
-    def __setattribute__(self, name, value):
-        raise self.error
-
-def import_lazy_fail(parent, name, error):
-    fail = _FailOnAccess(error)
-    try:
-        module = __import__(parent)
-    except ImportError:
-        return fail
-    return getattr(module, name, fail)
-
-error = ImportError('pyOpenSSL is not installed')
-crypto = import_lazy_fail('OpenSSL', 'crypto', error)
-ossl = import_lazy_fail('OpenSSL', 'SSL', error)
+try:
+    from OpenSSL import crypto
+    from OpenSSL import SSL as ossl
+except ImportError:
+    crypto = None
+    ossl = None
 
 __target__ = 'ssl'
 
@@ -60,7 +48,7 @@ _OPENSSL_ATTRS = dict(
 )
 
 # TODO maybe shrink __implements__ if pyOpenSSL unavailable?
-if not isinstance(ossl, _FailOnAccess):
+if ossl is not None:
     CERT_NONE = ossl.VERIFY_NONE
     CERT_REQUIRED = ossl.VERIFY_PEER | ossl.VERIFY_FAIL_IF_NO_PEER_CERT
 
@@ -80,6 +68,9 @@ HAS_NPN = False # TODO
 class SSLError(OSError):
     pass
 
+class SSLSysCallError(SSLError):
+    pass
+
 class SSLZeroReturnError(SSLError):
     pass
 
@@ -247,7 +238,7 @@ class SSLSocket(object):
             self._conn.set_connect_state() # FIXME does this override do_handshake_on_connect=False?
 
         if self.connected and self._do_handshake_on_connect:
-            _safe_ssl_call(False, self._conn, 'do_handshake')
+            self.do_handshake()
 
     @property
     def connected(self):
@@ -275,6 +266,17 @@ class SSLSocket(object):
         return _safe_ssl_call(self._suppress_ragged_eofs, self._conn, 'recv',
                                bufsize, flags)
 
+    def recv_into(self, buffer, bufsize=None, flags=None):
+        # A temporary recv_into implementation. Should be replaced when
+        # PyOpenSSL has merged pyca/pyopenssl#121.
+        if bufsize is None:
+            bufsize = len(buffer)
+
+        data = self.recv(bufsize, flags)
+        data_len = len(data)
+        buffer[0:data_len] = data
+        return data_len
+
     def send(self, data, flags=None):
         return _safe_ssl_call(False, self._conn, 'send', data, flags)
 
@@ -308,15 +310,19 @@ class SSLSocket(object):
         # The standard getpeercert() takes the nice X509 object tree returned
         # by OpenSSL and turns it into a dict according to some format it seems
         # to have made up on the spot. Here, we do our best to emulate that.
-        return dict(
+        result = dict(
             issuer=to_components(cert.get_issuer()),
             subject=to_components(cert.get_subject()),
             version=cert.get_subject(),
             serialNumber=cert.get_serial_number(),
             notBefore=cert.get_notBefore(),
             notAfter=cert.get_notAfter(),
-            subjectAltName=[('DNS', value) for value in get_subject_alt_name(cert)],
         )
+        if get_subject_alt_name is not None:
+            result.update(
+                subjectAltName=[('DNS', value) for value in get_subject_alt_name(cert)],
+            )
+        return result
 
     # FIXME support other arguments - not included in the signature to make
     # calls that expect them fail fast - use codecs.open()?
@@ -610,6 +616,7 @@ class SSLContext(object):
         self._ctx = ossl.Context(protocol)
         self.options = OP_ALL
         self.check_hostname = False
+        self.set_default_verify_paths()
 
     @property
     def options(self):
@@ -626,7 +633,6 @@ class SSLContext(object):
 
     @verify_mode.setter
     def verify_mode(self, value):
-        # TODO verify exception is raised on failure
         self._ctx.set_verify(value, lambda conn, cert, errnum, errdepth, ok: ok)
 
     def set_default_verify_paths(self):
@@ -648,6 +654,9 @@ class SSLContext(object):
             self._ctx.set_password_cb(lambda max_length, prompt_twice, userdata: password)
         self._ctx.use_privatekey_file(keyfile or certfile)
 
+    def set_ciphers(self, cipher_list):
+        return self._ctx.set_cipher_list(cipher_list)
+
     def set_npn_protocols(self, protocols):
         raise NotImplementedError() # TODO
 
@@ -661,14 +670,14 @@ class SSLContext(object):
 
 
 def wrap_socket(sock, keyfile=None, certfile=None, server_side=False,
-                cert_reqs=CERT_NONE, ssl_version=PROTOCOL_SSLv23, ca_certs=None,
+                cert_reqs=None, ssl_version=None, ca_certs=None,
                 do_handshake_on_connect=True, suppress_ragged_eofs=True,
                 ciphers=None):
     # TODO the stdlib docs say the SSLContext isn't constructed until connect()
     # is called on the socket, if it's not already connected. Check if we need
     # to emulate that behavior as well, or if it's just an optimization.
-    ctx = SSLContext(ssl_version)
-    ctx.verify_mode = cert_reqs
+    ctx = SSLContext(ssl_version if ssl_version is not None else PROTOCOL_SSLv23)
+    ctx.verify_mode = cert_reqs if cert_reqs is not None else CERT_NONE
     if certfile is not None:
         ctx.load_cert_chain(certfile, keyfile)
     if ca_certs is not None:
@@ -678,3 +687,8 @@ def wrap_socket(sock, keyfile=None, certfile=None, server_side=False,
     return ctx.wrap_socket(sock, server_side=server_side,
                            do_handshake_on_connect=do_handshake_on_connect,
                            suppress_ragged_eofs=suppress_ragged_eofs)
+
+
+if ossl is None:
+    for impl in ['SSLSocket', 'SSLContext', 'wrap_socket', '_fileobject']:
+        del locals()[impl]
